import logging

from twisted.application.service import IServiceMaker, Service, MultiService
from twisted.internet.endpoints import serverFromString
from twisted.internet.protocol import Factory
from twisted.internet.task import LoopingCall
from twisted.application.internet import TCPServer
from twisted.plugin import IPlugin
from twisted.python import usage, log
from twisted.web.client import Agent
from zope.interface import implementer

from bluefloodserver.protocols import MetricLineReceiver, MetricPickleReceiver
from bluefloodserver.collect import MetricCollection, ConsumeFlush, BluefloodFlush
from bluefloodserver.blueflood import BluefloodEndpoint

from txKeystone import KeystoneAgent


class Options(usage.Options):
    AUTH_URL = 'https://identity.api.rackspacecloud.com/v2.0/tokens'
    DEFAULT_TTL = 60 * 60 * 24
    optParameters = [
        ['endpoint',      'e', 'tcp:2004', 'Twisted formatted endpoint to listen to pickle protocol metrics'],
        ['interval',      'i', 30, 'Metric send interval, sec'],
        ['blueflood',     'b', 'http://localhost:19000', 'Blueflood server ingest URL (schema, host, port)'],
        ['tenant',        't', '', 'Blueflood tenant ID'],
        ['metric_prefix', 'p', '', 'prefix metric name with this string before sending them to Blueflood'],
        ['ttl',           '',  DEFAULT_TTL, 'TTL value for metrics, sec'],
        ['user',          'u', '', 'Rackspace authentication username. Leave empty if no auth is required'],
        ['key',           'k', '', 'Rackspace authentication password'],
        ['auth_url',      '',  AUTH_URL, 'Auth URL'],
        ['limit',         '',  0, 'Blueflood json payload limit, bytes. 0 means no limit']
    ]


class GraphiteMetricFactory(Factory):

    def __init__(self):
        flusher = ConsumeFlush()
        self._metric_collection = MetricCollection(flusher)

    def processMetric(self, metric, datapoint):
        log.msg('Receive metric {} {}:{}'.format(metric, datapoint[0], datapoint[1]), level=logging.DEBUG)
        self._metric_collection.collect(metric, datapoint)

    def flushMetric(self):
        try:
            log.msg('Sending {} metrics'.format(self._metric_collection.count()), level=logging.DEBUG)
            self._metric_collection.flush()
        except Exception, e:
            log.err(e)


class MetricService(Service):

    def __init__(self, **kwargs):
        self.protocol_cls = kwargs.get('protocol_cls')
        self.endpoint = kwargs.get('endpoint')
        self.flush_interval = kwargs.get('interval')
        self.blueflood_url = kwargs.get('blueflood_url')
        self.tenant = kwargs.get('tenant')
        self.ttl = kwargs.get('ttl')
        self.user = kwargs.get('user')
        self.key = kwargs.get('key')
        self.auth_url = kwargs.get('auth_url')
        self.limit = kwargs.get('limit', 0)
        self.metric_prefix = kwargs.get('metric_prefix', None)
        self.port = None

    def startService(self):
        from twisted.internet import reactor

        server = serverFromString(reactor, self.endpoint)
        log.msg('Start listening at {}'.format(self.endpoint))
        factory = GraphiteMetricFactory.forProtocol(self.protocol_cls)
        agent = Agent(reactor)
        if self.user:
            agent = KeystoneAgent(agent, self.auth_url, (self.user, self.key))
            log.msg('Auth URL: {}, user: {}'.format(self.auth_url, self.user))
        self._setup_blueflood(factory, agent)
        self.timer = LoopingCall(factory.flushMetric)
        self.timer.start(self.flush_interval)

        d = server.listen(factory)
        d.addCallback(self._store_listening_port)
        return d

    def _store_listening_port(self, port):
        self.port = port

    def stopService(self):
        if self.port:
            self.port.stopListening()
        self.timer.stop()

    def _setup_blueflood(self, factory, agent):
        log.msg('Send metrics to {} as tenant {} with {} sec interval'
            .format(self.blueflood_url, self.tenant, self.flush_interval))
        log.msg('Limit is {} bytes'.format(self.limit))
        client = BluefloodEndpoint(
            ingest_url=self.blueflood_url,
            tenant=self.tenant,
            agent=agent,
            limit=int(self.limit))
        flusher = BluefloodFlush(client=client, ttl=self.ttl, metric_prefix=self.metric_prefix)
        factory._metric_collection.flusher = flusher

@implementer(IServiceMaker, IPlugin)
class MetricServiceMaker(object):
    tapname = 'blueflood-forward'
    description = 'Forwarding metrics from graphite sources to Blueflood'
    options = Options

    def makeService(self, options):
        return MetricService(
            protocol_cls=MetricPickleReceiver,
            endpoint=options['endpoint'],
            interval=float(options['interval']),
            blueflood_url=options['blueflood'],
            tenant=options['tenant'],
            ttl=int(options['ttl']),
            user=options['user'],
            key=options['key'],
            auth_url=options['auth_url'],
            limit=options['limit'],
            metric_prefix=options['metric_prefix']
        )


serviceMaker = MetricServiceMaker()



'''
This is a simple test script that can be used to send pickle format 
metric to a twistd server running blueflood-forward. 
Modify the script as necessary for your testing need.
'''
import pickle
import socket
import struct
import time
import random

LISTEN_URL = ('localhost', 2004)
def send_to_socket(listOfMetricTuples):
    payload = pickle.dumps(listOfMetricTuples, protocol=2)
    header = struct.pack("!L", len(payload))
    message = header + payload
    print message
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect(LISTEN_URL)
    sock.sendall(message)
    sock.close()

timestamp = int(time.time())
value = random.randint(100, 200)
name = 'my.metric.name.2'
send_to_socket([(name, (timestamp, value))])



import time
import os
import mock

from bluefloodserver.blueflood import LimitExceededException
import bluefloodserver.collect as cl


def test_collect_and_flush():
    flusher = mock.MagicMock()
    collector = cl.MetricCollection(flusher)

    timestamp = time.time()
    metrics = []
    for i in range(10):
        metrics.append(('foo.bar.baz', i, timestamp))
        collector.collect('foo.bar.baz', (i, timestamp))
    collector.flush()

    assert flusher.flush.called_once_with(metrics)


def test_file_flish():
    file_name = 'output.txt'
    flusher = cl.FileFlush(file_name)
    collector = cl.MetricCollection(flusher)

    timestamp = time.time()
    for i in range(10):
        collector.collect('foo.bar.baz', (i, timestamp))
    collector.flush()

    file_data = open(file_name).read()
    os.remove(file_name)

    assert file_data == '\n'.join(
        ['foo.bar.baz {} {}'.format(i, timestamp) 
            for i in range(10)] + [''])

def test_blueflood_flush():
    client = mock.MagicMock()
    flusher = cl.BluefloodFlush(client)
    collector = cl.MetricCollection(flusher)

    timestamp = time.time()
    for i in range(10):
        collector.collect('foo.bar.baz', (i, timestamp))
    collector.flush()

    assert client.ingest.call_count == 10
    assert client.commit.called_once()

def test_blueflood_flush_with_limits():
    client = mock.MagicMock()
    client.ingest.side_effect = [LimitExceededException, None]
    flusher = cl.BluefloodFlush(client)
    collector = cl.MetricCollection(flusher)

    timestamp = time.time()
    collector.collect('foo.bar.baz', (0, timestamp))
    collector.flush()

    assert client.ingest.call_count == 2
    assert client.commit.call_count == 2



import pickle

import pytest
import mock
import bluefloodserver.protocols


def test_lineReceiver():
    handler = bluefloodserver.protocols.MetricLineReceiver()
    handler.factory = mock.MagicMock()
    handler.lineReceived('foo.bar.baz 22.2 124357823')
    handler.factory.processMetric.assert_called_once_with('foo.bar.baz', (124357823.0, 22.2))

def test_pickleReceiver():
    handler = bluefloodserver.protocols.MetricPickleReceiver()
    handler.factory = mock.MagicMock()
    timestamp = 12345678.0
    metrics = [('foo.bar.baz', (timestamp, i)) for i in range(5)]
    handler.connectionMade()
    handler.stringReceived(pickle.dumps(metrics))

    assert handler.factory.processMetric.call_count == 5
    handler.factory.processMetric.assert_called_with('foo.bar.baz', (timestamp, 4.0))

def test_datagramReceiver():
    handler = bluefloodserver.protocols.MetricDatagramReceiver()
    handler.factory = mock.MagicMock()
    timestamp = 12345678.0
    metrics = ['foo.bar.baz {} {}'.format(i, timestamp) for i in range(5)]
    handler.datagramReceived('\n'.join(metrics), ('localhost', 2004))

    assert handler.factory.processMetric.call_count == 5
    handler.factory.processMetric.assert_called_with('foo.bar.baz', (timestamp, 4.0))

    


import time

from twisted.internet import reactor
from twisted.internet.protocol import DatagramProtocol
from twisted.internet.error import ConnectionDone
from twisted.protocols.basic import LineOnlyReceiver, Int32StringReceiver

import util


class MetricReceiver:

    def connectionMade(self):
      pass

    def metricReceived(self, metric, datapoint):
        if datapoint[1] != datapoint[1]:
            return
        if int(datapoint[0]) == -1:
            datapoint = (time.time(), datapoint[1])

        self.factory.processMetric(metric, datapoint)

class MetricLineReceiver(MetricReceiver, LineOnlyReceiver):
    delimiter = '\n'

    def lineReceived(self, line):
        try:
            metric, value, timestamp = line.strip().split()
            datapoint = (float(timestamp), float(value))
        except ValueError:
            return

        self.metricReceived(metric, datapoint)


class MetricDatagramReceiver(MetricReceiver, DatagramProtocol):

    def datagramReceived(self, data, (host, port)):
        for line in data.splitlines():
            try:
                metric, value, timestamp = line.strip().split()
                datapoint = (float(timestamp), float(value))

                self.metricReceived(metric, datapoint)
            except ValueError:
                pass


class MetricPickleReceiver(MetricReceiver, Int32StringReceiver):
    MAX_LENGTH = 2 ** 20

    def connectionMade(self):
        MetricReceiver.connectionMade(self)
        self.unpickler = util.get_unpickler(
            insecure=True)

    def stringReceived(self, data):
        try:
            datapoints = self.unpickler.loads(data)
        except util.pickle.UnpicklingError:
            return

        for (metric, datapoint) in datapoints:
            try:
                datapoint = (float(datapoint[0]), float(datapoint[1]))
            except ValueError:
                continue

            self.metricReceived(metric, datapoint)



#!/usr/bin/env python

import urllib2
import urlparse
import json

from StringIO import StringIO

from twisted.internet import reactor
from twisted.internet.defer import Deferred, inlineCallbacks, returnValue
from twisted.internet.protocol import Protocol
from twisted.web.client import Agent, FileBodyProducer, readBody
from twisted.web.http_headers import Headers


def _get_metrics_url(url, tenantId):
    return url + '/v2.0/'\
        + tenantId + '/ingest'

def _get_metrics_query_url(url, tenantId,
                           metricName, start, end, points):
    return url + '/v2.0/' + tenantId\
        + '/views/' + metricName\
        + '?from=' + str(start) + '&to=' + str(end) + '&points=' + str(points)

def _get_metrics_query_url_resolution(url, tenantId,
                           metricName, start, end, resolution='FULL'):
    return url + '/v2.0/' + tenantId\
        + '/views/' + metricName\
        + '?from=' + str(start) + '&to=' + str(end) + '&resolution=' + resolution


class LimitExceededException(Exception):
    pass


class BluefloodEndpoint():

    def __init__(self, ingest_url='http://localhost:19000', retrieve_url='http://localhost:20000', tenant='tenant-id', agent=None, limit=None):
        self.agent = agent
        self.ingest_url = ingest_url
        self.retrieve_url = retrieve_url
        self.tenant = tenant
        self._json_buffer = []
        self.headers = {}
        self.limit = limit
        self._buffer_size = 0

    def ingest(self, metric_name, time, value, ttl):
        if not isinstance(time, list):
            time = [time]
        if not isinstance(value, list):
            value = [value]
        if len(time) != len(value):
            raise Exception('time and value list lengths differ')


        data = [{
            "collectionTime": t*1000,
            "ttlInSeconds": ttl,
            "metricValue": v,
            "metricName": metric_name
        } for t,v in zip(time, value)]
        row_size = len(json.dumps(data))
        # len('[]'') = 2, len(',' * total) = len(data) - 1
        if self.limit and row_size + len(self._json_buffer) - 1 + self._buffer_size > self.limit:
            raise LimitExceededException('JSON limit exceeded. Commit metrics before next ingest')
        self._json_buffer.extend(data)
        self._buffer_size += row_size

    @inlineCallbacks
    def commit(self):
        body = FileBodyProducer(StringIO(json.dumps(self._json_buffer)))
        d = self.agent.request(
            'POST',
            _get_metrics_url(self.ingest_url, self.tenant),
            Headers(self.headers),
            body)

        resp = yield d
        if resp.code == 200:
            self._json_buffer = []
            self._buffer_size = 0
        returnValue(None)


    @inlineCallbacks
    def retrieve_points(self, metric_name, start, to, points):
        d = self.agent.request(
            'GET',
            _get_metrics_query_url(self.retrieve_url,
                self.tenant, metric_name, start, to, points),
            Headers(self.headers),
            None)
        resp = yield d
        body = yield readBody(resp)

        returnValue(json.loads(body))

    retrieve = retrieve_points

    @inlineCallbacks
    def retrieve_resolution(self, metric_name, start, to, resolution='FULL'):
        d = self.agent.request(
            'GET',
            _get_metrics_query_url_resolution(self.retrieve_url,
                self.tenant, metric_name, start, to, resolution),
            Headers(self.headers),
            None)

        resp = yield d
        body = yield readBody(resp)

        returnValue(json.loads(body))



#!/usr/bin/env python

import os
import json
import pickle
import pytest
import SocketServer
import socket
import struct
import subprocess
import threading
import time
import urllib2

import http_server_mock

from twisted.internet import reactor, endpoints
from twisted.web import server


b_handler = None
a_handler = None
t = None


TENANT = 'tenant'
AUTH_URL = 'http://localhost:8001'
AUTH_USER = 'user1'
AUTH_KEY = 'key2'
BLUEFLOOD_URL = 'http://localhost:8000'
LISTEN_URL = ('localhost', 2004)
INTERVAL = 5
LIMIT = 1024 # 1 kilobyte

def run_twisted(user=None, key=None):
    args = ['twistd', '-n', '-l', 'log.txt', 'blueflood-forward',
     '-b', BLUEFLOOD_URL, '-i', str(INTERVAL), '-t', TENANT, '--limit', str(LIMIT)]
    if user:
        args.extend(['-u', user, '-k', key, '--auth_url', AUTH_URL])
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return p

def send_to_socket(listOfMetricTuples):
    payload = pickle.dumps(listOfMetricTuples, protocol=2)
    header = struct.pack("!L", len(payload))
    message = header + payload
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect(LISTEN_URL)
    sock.sendall(message)
    sock.close()

def setup_module(module):
    """ setup any state specific to the execution of the given module."""
    
    global b_handler, a_handler, t
    # blueflood server mock
    b_handler = http_server_mock.MockServerHandler()
    # auth server mock
    a_handler = http_server_mock.MockServerHandler()

    endpoints.serverFromString(reactor, "tcp:8000").listen(server.Site(b_handler))
    endpoints.serverFromString(reactor, "tcp:8001").listen(server.Site(a_handler))

    t = threading.Thread(target=reactor.run, args=(0, ))
    t.daemon = True
    t.start()

def teardown_module(module):
    """ teardown any state that was previously setup with a setup_module
    method.
    """
    global t

    reactor.stop()
    t = None


@pytest.fixture
def twistd(request):
    return fixture_run_twisted(request, None, None)

@pytest.fixture
def twistd_auth(request):
    return fixture_run_twisted(request, AUTH_USER, AUTH_KEY)
    
def fixture_run_twisted(request, user, key):
    p = run_twisted(user, key)
    def fin():
        p.terminate()
        print 'Waiting process to terminate'
        watchdog_timer = threading.Timer(5.0, p.kill, [p])
        watchdog_timer.start()
        p.wait()
        watchdog_timer.cancel()
        print 'quit', p, time.time()
        assert p.returncode == 0
    request.addfinalizer(fin)
    return p

def setup_and_teardown_handler(request, handler):
    cv = threading.Condition()
    handler.cv = cv
    def fin():
        handler.reset_to_default_reply()
        handler.cv = None
        handler.data = []
    request.addfinalizer(fin)
    return handler

@pytest.fixture
def blueflood_handler(request):
    global b_handler
    return setup_and_teardown_handler(request, b_handler)

@pytest.fixture
def auth_handler(request):
    global a_handler
    return setup_and_teardown_handler(request, a_handler)

def test_noauth(blueflood_handler, twistd):
    blueflood_handler.should_reply_forever(200, '', {})

    time.sleep(1.5)
    cv = blueflood_handler.cv
    timestamp = int(time.time())
    value = 10.0
    name = 'foo.bar.baz'
    send_to_socket([(name, (timestamp, value))])
    with cv:
        cv.wait(INTERVAL * 2)

    assert len(blueflood_handler.data) == 1
    _, headers, path, data = blueflood_handler.data[0]
    assert path == '/v2.0/tenant/ingest'
    metrics = json.loads(data)
    assert len(metrics) == 1
    assert metrics[0]['metricValue'] == value
    assert int(metrics[0]['collectionTime']/1000) == timestamp
    assert metrics[0]['metricName'] == name

def test_auth(blueflood_handler, auth_handler, twistd_auth):
    response = """{"access":{"token":{"id":"eb5e1d9287054898a55439137ea68675","expires":"2014-12-14T22:54:49.574Z","tenant":{"id":"836986","name":"836986"}}}}"""

    auth_handler.should_reply_forever(200, response, {})
    blueflood_handler.should_reply_once(200, '', {})
    time.sleep(1.0)

    timestamp = int(time.time())
    value = 10.0
    name = 'foo.bar.baz'
    send_to_socket([(name, (timestamp, value))])

    with auth_handler.cv:
        auth_handler.cv.wait(INTERVAL)
    assert len(auth_handler.data) == 1

    with blueflood_handler.cv:
        blueflood_handler.cv.wait(INTERVAL)
    assert len(blueflood_handler.data) == 1
    headers = blueflood_handler.data[0][1]
    assert 'x-auth-token' in headers
    assert headers['x-auth-token'] == 'eb5e1d9287054898a55439137ea68675'

def test_splitting_payload(blueflood_handler, twistd):
    blueflood_handler.should_reply_forever(200, '', {})
    cv = blueflood_handler.cv

    time.sleep(1.5)
    timestamp = int(time.time())
    value = 10.0
    name = 'foo.bar.baz'
    metrics = {
        'collectionTime': timestamp,
        'metricName': name,
        'metricValue': value,
        'ttlInSeconds': 86400
    }
    metric_size = len(json.dumps(metrics))
    count = LIMIT / metric_size + 1
    send_to_socket([(name, (timestamp, value)) for i in range(count)])
    with cv:
        cv.wait(INTERVAL * 2)

    assert len(blueflood_handler.data) > 1
    _, _, _, data = blueflood_handler.data[0]
    assert len(data) < LIMIT
    _, _, _, data_rest = blueflood_handler.data[1]
    assert len(data) != len(data_rest)


#!/usr/bin/env python
import urllib2
import time
import json 
import pytest

from decorator import decorator

from twisted.web.client import Agent, FileBodyProducer, readBody
from twisted.internet import reactor, defer
from bluefloodserver.blueflood import BluefloodEndpoint, LimitExceededException

# python-twisted source code
# TODO: find better way to enable twisted plugin for this particular test module
# tried pytest_plugins directly here with no luck
@decorator
def inlineCallbacks(fun, *args, **kw):
    return defer.inlineCallbacks(fun)(*args, **kw)

try:
    b = BluefloodEndpoint()
    data = urllib2.urlopen(b.ingest_url).read()
except urllib2.HTTPError, e:
    skip = False
except urllib2.URLError, e:
    skip = True
else:
    skip = False

@pytest.fixture
def setup():
    return BluefloodEndpoint(agent=Agent(reactor))

@inlineCallbacks
@pytest.mark.skipif(skip, reason="Blueflood isn't running")
def testSingleIngest(setup):
    endpoint = setup
    name = 'example.metric.single.ingest'
    ttl = 10
    endpoint.ingest(name, 1376509892612, 50, ttl) 
    yield endpoint.commit()

@inlineCallbacks
@pytest.mark.skipif(skip, reason="Blueflood isn't running")
def testListIngest(setup):
    endpoint = setup
    name = 'example.metric.list.ingest'
    ttl = 10
    endpoint.ingest(name,
                    [1376509892612, 1376509892613, 1376509892614], 
                    [50, 51, 52], 
                    ttl)
    yield endpoint.commit()
    with pytest.raises(Exception):
        endpoint.ingest(name, 
                        [1376509892612, 1376509892613, 1376509892614], 
                        50, 
                        ttl)

@inlineCallbacks
@pytest.mark.skipif(skip, reason="Blueflood isn't running")
def testMultipleIngest(setup):
    endpoint = setup
    name = 'example.metric.multiple.ingest'
    ttl = 10
    times, values = list(range(5)), [50.0] * 5
    for t, v in zip(times, values):
         endpoint.ingest(name, t, v, ttl)
    yield endpoint.commit()
    data = yield endpoint.retrieve_resolution(name, 0, 10)
    assert len(data['values']) == 5

@inlineCallbacks
@pytest.mark.skipif(skip, reason="Blueflood isn't running")
def testRetrieveByResolution(setup):
    endpoint = setup
    name = 'example.metric.retrieve'
    ttl = 10
    timestamp = int(time.time())
    value = 50
    # insert first
    endpoint.ingest(name, timestamp, value, ttl)
    yield endpoint.commit()
    # range is 0-time
    data = yield endpoint.retrieve_resolution(name, 0, timestamp + 10)
    assert len(data) != 0
    assert len(data['values']) != 0
    assert data['values'][0]['numPoints'] == 1
    assert data['values'][0]['average'] == value

@inlineCallbacks
@pytest.mark.skipif(skip, reason="Blueflood isn't running")
def testRetrieveByPoints(setup):
    endpoint = setup
    name = 'example.metric.retrieve'
    ttl = 10
    timestamp = 1234567
    value = 50
    # insert first
    endpoint.ingest(name, timestamp, value, ttl)
    yield endpoint.commit()
    # range is 0-time
    data = yield endpoint.retrieve_points(name, 0, timestamp, 200)
    assert len(data) != 0
    assert len(data['values']) != 0
    assert data['values'][0]['numPoints'] == 1
    assert data['values'][0]['average'] == value

@inlineCallbacks
def testNoConnection():
    endpoint = BluefloodEndpoint(ingest_url='http://localhost:9623',
        retrieve_url='http://localhost:8231')
    with pytest.raises(Exception):
        d = yield endpoint.commit()
    with pytest.raises(Exception):
        yield endpoint.retrieve_points('', 0, 0, 0)
    with pytest.raises(Exception):
        yield endpoint.retrieve_resolution('', 0, 0)

def testLimitIngestion():
    timestamp = int(time.time())
    name = ''
    value = 1.0
    ttl = 0
    metric = {
        'collectionTime':timestamp,
        'metricName': name,
        'metricValue': value,
        'ttlInSeconds': ttl
    }
    metric_size = len(json.dumps(metric))
    count = 5
    endpoint = BluefloodEndpoint(limit=metric_size*count)
    for i in range(count-1):
        endpoint.ingest(name, timestamp, value, ttl)
    with pytest.raises(LimitExceededException):
        endpoint.ingest(name, timestamp, value, ttl)


if __name__ == '__main__':
    pytest.main()



import StringIO
import json
import mock
import pytest
import twisted.plugins.graphite_blueflood_plugin as plugin
from twisted.web.client import Agent
from txKeystone import KeystoneAgent
from twisted.test import proto_helpers


def test_service():
    service = plugin.serviceMaker.makeService(plugin.Options())
    assert isinstance(service, plugin.MetricService)

def test_service_simple_agent():
    options = plugin.Options()
    service = plugin.serviceMaker.makeService(options)
    service._setup_blueflood = mock.MagicMock()
    service.startService()
    assert service._setup_blueflood.called
    assert isinstance(service._setup_blueflood.call_args_list[0][0][1], Agent)
    service.stopService()

def test_service_auth_agent():
    options = plugin.Options()
    options.parseOptions(['--user', 'user', '--key', 'key'])
    service = plugin.serviceMaker.makeService(options)
    service._setup_blueflood = mock.MagicMock()
    service.startService()
    assert service._setup_blueflood.called
    assert isinstance(service._setup_blueflood.call_args_list[0][0][1], KeystoneAgent)
    service.stopService()


def test_factory():
    factory = plugin.GraphiteMetricFactory()
    factory.protocol = plugin.MetricLineReceiver
    factory._metric_collection = mock.MagicMock()
    proto = factory.buildProtocol(('127.0.0.1', 0))
    tr = proto_helpers.StringTransport()
    proto.makeConnection(tr)

    proto.dataReceived('foo.bar.baz 123 123456789.0\n')
    assert factory._metric_collection.collect.called_once_with('foo.bar.baz', 123456789.0, 123.0)

def test_send_blueflood():
    factory = plugin.GraphiteMetricFactory()
    agent = mock.MagicMock()
    factory.protocol = plugin.MetricLineReceiver
    plugin.MetricService(
        protocol_cls=factory.protocol,
        endpoint='',
        interval=5,
        blueflood_url='http://bluefloodurl:190',
        tenant='tenant',
        ttl=30)._setup_blueflood(factory, agent)

    proto = factory.buildProtocol(('127.0.0.1', 0))
    tr = proto_helpers.StringTransport()
    proto.makeConnection(tr)

    proto.dataReceived('foo.bar.baz 123 123456789.0\n')
    factory.flushMetric()
    assert agent.request.called 
    assert len(agent.request.call_args_list) == 1
    rq = agent.request.call_args_list[0][0]
    assert rq[0] == 'POST'
    assert rq[1] == 'http://bluefloodurl:190/v2.0/tenant/ingest'
    metrics = json.loads(rq[3]._inputFile.read())
    assert len(metrics) == 1
    assert metrics[0]['metricName'] == 'foo.bar.baz'
    assert metrics[0]['metricValue'] == 123.0



import copy
import os
import pwd
import sys

from os.path import abspath, basename, dirname
try:
  from cStringIO import StringIO
except ImportError:
  from StringIO import StringIO
try:
  import cPickle as pickle
  USING_CPICKLE = True
except ImportError:
  import pickle
  USING_CPICKLE = False

from time import sleep, time
from twisted.python.util import initgroups
from twisted.scripts.twistd import runApp


# This whole song & dance is due to pickle being insecure
# yet performance critical for carbon. We leave the insecure
# mode (which is faster) as an option (USE_INSECURE_UNPICKLER).
# The SafeUnpickler classes were largely derived from
# http://nadiana.com/python-pickle-insecure
if USING_CPICKLE:
  class SafeUnpickler(object):
    PICKLE_SAFE = {
      'copy_reg': set(['_reconstructor']),
      '__builtin__': set(['object']),
    }

    @classmethod
    def find_class(cls, module, name):
      if not module in cls.PICKLE_SAFE:
        raise pickle.UnpicklingError('Attempting to unpickle unsafe module %s' % module)
      __import__(module)
      mod = sys.modules[module]
      if not name in cls.PICKLE_SAFE[module]:
        raise pickle.UnpicklingError('Attempting to unpickle unsafe class %s' % name)
      return getattr(mod, name)

    @classmethod
    def loads(cls, pickle_string):
      pickle_obj = pickle.Unpickler(StringIO(pickle_string))
      pickle_obj.find_global = cls.find_class
      return pickle_obj.load()

else:
  class SafeUnpickler(pickle.Unpickler):
    PICKLE_SAFE = {
      'copy_reg': set(['_reconstructor']),
      '__builtin__': set(['object']),
    }

    def find_class(self, module, name):
      if not module in self.PICKLE_SAFE:
        raise pickle.UnpicklingError('Attempting to unpickle unsafe module %s' % module)
      __import__(module)
      mod = sys.modules[module]
      if not name in self.PICKLE_SAFE[module]:
        raise pickle.UnpicklingError('Attempting to unpickle unsafe class %s' % name)
      return getattr(mod, name)

    @classmethod
    def loads(cls, pickle_string):
      return cls(StringIO(pickle_string)).load()


def get_unpickler(insecure=False):
  if insecure:
    return pickle
  else:
    return SafeUnpickler



#!/usr/bin/env python

import sys
import time
from twisted.web import resource, server
from twisted.internet import reactor, endpoints


class MockServerHandler(resource.Resource):
    isLeaf = True

    def __init__(self, cv=None):
        resource.Resource.__init__(self)
        self.cv = cv
        self.data = []
        self.reply = []
        self.default_reply = (200, '', {})

    def should_reply_once(self, status, response, headers):
        self.reply.append((status, response, headers))

    def should_reply_forever(self, status, response, headers):
        self.default_reply = (status, response, headers)

    def reset_to_default_reply(self):
        self.reply = []

    def _get_next_reply(self):
        code, response, headers = self.reply[0] if self.reply else self.default_reply
        try:
            del self.reply[0]
        except IndexError:
            pass
        return code, response, headers


    def render_POST(self, request):
        data = request.content.read()
        self.data += [(time.time(), dict(request.received_headers), request.path, data)]
        print self.data[-1]
        print ''
        if self.cv:
            with self.cv:
                self.cv.notify()

        code, response, headers = self._get_next_reply()
        request.setResponseCode(code)
        for key, value in headers.items():
            request.setHeader(key, value)
        return response

if __name__ == '__main__':
    b_handler = MockServerHandler()
    a_handler = MockServerHandler()
    response = """{"access":{"token":{"id":"eb5e1d9287054898a55439137ea68675","expires":"2014-12-14T22:54:49.574Z","tenant":{"id":"836986","name":"836986"}}}}"""
    endpoints.serverFromString(reactor, "tcp:8000").listen(server.Site(b_handler))
    endpoints.serverFromString(reactor, "tcp:8001").listen(server.Site(a_handler))
    a_handler.should_reply_forever(200, response, {})
    reactor.run()



import os

from twisted.logger import FilteringLogObserver, LogLevelFilterPredicate, LogLevel, jsonFileLogObserver
from twisted.python import logfile

log_dir = os.environ.get("LOG_DIR", '/var/log/')
log_level = os.environ.get("TWISTED_LOG_LEVEL", 'INFO').lower()
log_rotate_length = int(os.environ.get("LOG_ROTATE_LENGTH", 100000000))
max_rotated_log_files = int(os.environ.get("MAX_LOG_ROTATED_FILES", 10))

def get_log_observer():
    f = logfile.LogFile("carbon_forwarder.log", log_dir, rotateLength=log_rotate_length, maxRotatedFiles=max_rotated_log_files)
    observer = jsonFileLogObserver(f)
    filterer = FilteringLogObserver(observer,
        [LogLevelFilterPredicate(
            LogLevel.levelWithName(log_level))])
    return filterer



from __future__ import print_function
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import io
import codecs
import os
import sys

import bluefloodserver
import carbonforwarderlogging

here = os.path.abspath(os.path.dirname(__file__))

def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)

long_description = read('README.md')

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)

setup(
    name='graphite-blueflood',
    version="0.2.1",
    url='https://github.com/rackerlabs/blueflood-carbon-forwarder',
    license='Apache Software License',
    author='Rackspace Metrics',
    tests_require=['pytest', 'mock', 'twisted', 'pytest-twisted'],
    install_requires=['twisted', 'txKeystone'],
    cmdclass={'test': PyTest},
    author_email='cloudMetrics-dev@lists.rackspace.com',
    description='Sending graphite metrics to blueflood',
    long_description=long_description,
    packages=find_packages(exclude=[]) + ['bluefloodserver'] + ['carbonforwarderlogging'] + ["twisted.plugins"],
    include_package_data=True,
    platforms='any',
    test_suite='tests',
    classifiers = [
        'Programming Language :: Python',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        ],
    extras_require={
        'testing': ['pytest'],
    }
)



import logging

from blueflood import LimitExceededException
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.python import log

class IFlush:

    def flush(self, metrics):
        raise('Not Implemented')

class ConsumeFlush(IFlush):

    def flush(self, metrics):
        pass

class FileFlush(IFlush):

    def __init__(self, filename):
        self.filename = filename

    def flush(self, metrics):
        with open(self.filename, 'a') as outfile:
            for name, time, value in metrics:
                outfile.write('{} {} {}\n'.format(name, time, value))


class BluefloodFlush(IFlush):

    def __init__(self, client, ttl=60*60*24, metric_prefix=None):
        self.client = client
        self.ttl = ttl
        self.metric_prefix = metric_prefix
        log.msg('Prepending all metrics with custom string {}'.format(metric_prefix), level=logging.DEBUG)

    @inlineCallbacks
    def flush(self, metrics):
        for name, time, value in metrics:
            if self.metric_prefix:
                metric_name = self.metric_prefix + '.' + name
            else:
                metric_name = name
            try:
                self.client.ingest(metric_name, time, value, self.ttl)
            except LimitExceededException:
                yield self.client.commit()
                self.client.ingest(metric_name, time, value, self.ttl)
        yield self.client.commit()
        returnValue(None)


class MetricCollection:        
    
    def __init__(self, flusher):
        self._metrics = []
        self.flusher = flusher

    def collect(self, metric, datapoint):
        self._metrics.append((metric, datapoint[0], datapoint[1]))

    def flush(self):
        if self._metrics:
            self.flusher.flush(self._metrics)
            self._metrics = []

    def count(self):
        return len(self._metrics)
